#region usings
using System;
using System.ComponentModel.Composition;
using System.Collections.Generic;
using System.Linq;

using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;
using VVVV.Core.Logging;

#endregion usings

namespace VVVV.Nodes
{
    #region PluginInfo
    [PluginInfo(Name = "midiLogger", Category = "Midi", Version = "0.1", Help = "Logs Midi Notes", Tags = "")]
    #endregion PluginInfo
    public class C0_1MidimidiLoggerNode : IPluginEvaluate
    {
        #region fields & pins
        [Input("Message", DefaultValue = 1.0, IsSingle = true)]
        public ISpread<int> FIMessage;

    	[Input("Data1", DefaultValue = 1.0, IsSingle = true)]
        public ISpread<int> FIData1;
    	
    	[Input("Data2", DefaultValue = 1.0, IsSingle = true)]
        public ISpread<int> FIData2;
    	
    	[Input("Timestamp", DefaultValue = 1.0, IsSingle = true)]
        public ISpread<int> FITimestamp;
    	
    	[Input("Reset", IsBang = true, IsSingle = true)]
		public ISpread<bool> FIReset;
    	
        [Output("Note")]
        public ISpread<int> FONote;

        [Output("Velocity")]
        public ISpread<int> FOVelocity;

    	[Output("Time")]
        public ISpread<int> FOTime;

    	[Output("Length")]
        public ISpread<int> FOLength;

        [Output("Channel")]
        public ISpread<int> FOChannel;


        [Import()]
        public ILogger FLogger;
        #endregion fields & pins

        private List<MidiShort> openNoteBuffer = new List<MidiShort>();
        private List<MidiShortOut> loggedNote = new List<MidiShortOut>();
    	
    	bool firstRun = true;
    	
        public void Evaluate(int SpreadMax)
        {	
	
            MidiShort inputMidi = new MidiShort
            {
                Message = FIMessage[0],
                Data1 = FIData1[0],
                Data2 = FIData2[0],
                Timestamp = FITimestamp[0]
            };


            if (openNoteBuffer.Count == 0 && isNoteNew(inputMidi)) addNote2Buffer();

            if (openNoteBuffer.Count != 0)
            {
                if (isNoteNew(inputMidi) && !isNoteEqual(openNoteBuffer.Last(), inputMidi)){
                   	addNote2Buffer();
                	
                	//FLogger.Log(LogType.Debug, "New Note add to Buffer - Count: {0}", openNoteBuffer.Count.ToString());  	
                }
            }

           if(openNoteBuffer.Count > 0 && isNoteEnd(inputMidi) && !isNoteEqual(openNoteBuffer.Last(), inputMidi))
            {
                MidiShort item = openNoteBuffer.Find(x => x.Data1 == inputMidi.Data1);
                if (item != null)
                {
                    loggedNote.Add(new MidiShortOut
                    {
                        Note = item.Data1,
                        Velocity = item.Data2,
                        Time = item.Timestamp,
                        Length = FITimestamp[0] - item.Timestamp,
                        Channel = item.Message - 144
                    });

                    openNoteBuffer.Remove(item);
                }


            int listLength = loggedNote.Count;

            if (listLength > 0)
	            {
	                FONote.SliceCount = listLength;
	            	FOVelocity.SliceCount = listLength;
	                FOTime.SliceCount = listLength;
	            	FOLength.SliceCount = listLength;
	                FOChannel.SliceCount = listLength;
	
	
	                for (int i = 0; i < listLength; i++)
	                {
	                    FONote[i] = loggedNote[i].Note;
	                    FOVelocity[i] = loggedNote[i].Velocity;
	                    FOTime[i] = loggedNote[i].Time;
	                    FOLength[i] = loggedNote[i].Length;
	                    FOChannel[i] = loggedNote[i].Channel;
	                }
	            }
	        }
        	
        	if(FIReset[0] || firstRun){
        		firstRun = false;
				reset();   
	        	//FLogger.Log(LogType.Debug, "New Note add to Buffer");
	        	}
        }
		
    	
        private bool isNoteEqual(MidiShort lastMidiShort, MidiShort inMidiShort)
        {
            if (lastMidiShort.Data2 != inMidiShort.Data2) return false;
            if (lastMidiShort.Message != inMidiShort.Message) return false;

            return true;
        }
   	
    	private void addNote2Buffer()
        {
            openNoteBuffer.Add(new MidiShort
            {
                Message = FIMessage[0],
                Data1 = FIData1[0],
                Data2 = FIData2[0],
                Timestamp = FITimestamp[0]
            });
        }

    	private void reset()
    	{
    			loggedNote.Clear();
				openNoteBuffer.Clear();
   
                FONote.SliceCount = 0;
                FOVelocity.SliceCount = 0;
                FOTime.SliceCount = 0;
                FOLength.SliceCount = 0;
    		 	FOChannel.SliceCount = 0;
    	}

        private bool isNoteEnd(MidiShort inMidiShort)
        {
            if (inMidiShort.Data2 == 0) return true;
            if (128 <= inMidiShort.Message && inMidiShort.Message <= 143) return true;
            return false;
        }

        private bool isNoteNew(MidiShort inMidiShort)
        {
            //if (inMidiShort.Data2 != 0) return true;
            if (144 <= inMidiShort.Message && inMidiShort.Message <= 159 && inMidiShort.Data2 != 0) return true;
        	
            return false;
        }

        private class MidiShort
        {
            public int Message { get; set; }
            public int Data1 { get; set; }
            public int Data2 { get; set; }
            public int Timestamp { get; set; }
        }

        private class MidiShortOut
        {
            public int Note { get; set; }
            public int Velocity { get; set; }
            public int Time { get; set; }
            public int Length { get; set; }
            public int Channel { get; set; }
        }


    }
}
